# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import functools
import pathlib
import re

from dataclasses import dataclass
from typing import ClassVar, Optional, cast

import imageio.plugins.freeimage
import imageio.v3
import numpy as np
import numpy.typing as npt

from no_vtf.image import Image, ImageData, ImageDataTypes
from no_vtf.image.modifier.fp_precision_modifier import FPPrecisionModifier
from no_vtf.image.modifier.modifier import ImageModifier
from no_vtf.io.io import IO


@dataclass(frozen=True, kw_only=True)
class ImageIO(IO[Image[ImageDataTypes]]):
    format: str

    compress: Optional[bool] = None

    @classmethod
    def _initialize(cls) -> None:
        # download() seems to be untyped because of implicit reexport
        imageio.plugins.freeimage.download()  # type: ignore[no-untyped-call]

    _format_pattern: ClassVar[re.Pattern[str]] = re.compile(r"[a-z0-9]+", re.ASCII | re.IGNORECASE)

    def __post_init__(self) -> None:
        assert self._is_initialized(), "IO.initialize() must be called early"

        if not self._format_pattern.fullmatch(self.format):
            raise RuntimeError(f"Invalid format: {self.format}")

    def write(self, path: pathlib.Path, image: Image[ImageDataTypes]) -> None:
        backend = self._get_backend()
        backend.write(path, image)

    def readback(self, path: pathlib.Path, image: Image[ImageDataTypes]) -> None:
        backend = self._get_backend()
        backend.readback(path, image)

    def _get_backend(self) -> _ImageIOBackend:
        compress = self.compress
        if compress is None:
            compress = True

        extension = f".{self.format}"

        backend: _ImageIOBackend
        match self.format.lower():
            case "exr":
                backend = _ImageIOExrBackend(compress=compress)
            case "png":
                backend = _ImageIOPngBackend(compress=compress)
            case "targa" | "tga":
                backend = _ImageIOTgaBackend(compress=compress)
            case "tiff":
                backend = _ImageIOTiffBackend(compress=compress)
            case _:
                backend = _ImageIOBackend(extension=extension)
        return backend


class _ImageIOBackend:
    def __init__(
        self, *, imageio_format: Optional[str] = None, extension: Optional[str] = None
    ) -> None:
        self._imageio_format = imageio_format
        self._extension = extension

    def write(self, path: pathlib.Path, image: Image[ImageDataTypes]) -> None:
        kwargs = self._get_writer_kwargs(image)
        image = self._postprocess(image)
        data = self._get_data(image)

        legacy_opener = functools.partial(imageio.v3.imopen, legacy_mode=True)
        with legacy_opener(
            path, "w", plugin=self._imageio_format, extension=self._extension
        ) as image_resource:
            image_resource.write(data, **kwargs)

    def readback(self, path: pathlib.Path, image: Image[ImageDataTypes]) -> None:
        image = self._postprocess(image)
        data = self._get_data(image)

        legacy_opener = functools.partial(imageio.v3.imopen, legacy_mode=True)
        with legacy_opener(
            path, "r", plugin=self._imageio_format, extension=self._extension
        ) as image_resource:
            read_data = image_resource.read(index=0)

            if data.dtype != read_data.dtype:
                raise RuntimeError(f"{path!r}: Data type differs from what is in the file")

            if not self._compare_data(data, read_data):
                raise RuntimeError(f"{path!r}: Data differs from what is in the file")

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        return {}

    def _postprocess(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypes]:
        return image

    def _get_data(self, image: Image[ImageDataTypes]) -> ImageData:
        data = image.data()

        # write luminance into three channels when alpha is present
        if image.channels == "la":
            l_uint8: npt.NDArray[ImageDataTypes] = data[:, :, [0]]
            a_uint8: npt.NDArray[ImageDataTypes] = data[:, :, [1]]
            data = np.dstack((l_uint8, l_uint8, l_uint8, a_uint8))

        # remove last axis if its length is 1
        if data.shape[-1] == 1:
            data = data[..., 0]

        return data

    def _compare_data(self, data: ImageData, read_data: ImageData) -> bool:
        return np.array_equal(data, read_data)


class _ImageIOFreeImageBackend(_ImageIOBackend):
    IO_FLAGS: ClassVar = imageio.plugins.freeimage.IO_FLAGS

    def __init__(self, *, imageio_format: str, extension: str) -> None:
        super().__init__(imageio_format=imageio_format, extension=extension)

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        kwargs: dict[str, object] = {}
        kwargs["flags"] = self._get_flags(image)
        return kwargs

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        return 0


class _ImageIOExrBackend(_ImageIOFreeImageBackend):
    _fp_force_32_bits: ClassVar[
        ImageModifier[ImageDataTypes, ImageDataTypes]
    ] = FPPrecisionModifier(min=32, max=32)

    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="EXR-FI", extension=".exr")
        self.compress = compress

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        flags = 0
        flags |= self.IO_FLAGS.EXR_ZIP if self.compress else self.IO_FLAGS.EXR_NONE
        if not np.issubdtype(image.data().dtype, np.float16):
            flags |= self.IO_FLAGS.EXR_FLOAT
        return flags

    def _postprocess(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypes]:
        return self._fp_force_32_bits(image)


class _ImageIOPngBackend(_ImageIOFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="PNG-FI", extension=".png")
        self.compress = compress

    def _get_writer_kwargs(self, image: Image[ImageDataTypes]) -> dict[str, object]:
        kwargs: dict[str, object] = super()._get_writer_kwargs(image)
        kwargs["compression"] = 1 if self.compress else 0
        return kwargs

    def _compare_data(self, data: ImageData, read_data: ImageData) -> bool:
        if np.issubdtype(data.dtype, np.uint8) and np.issubdtype(read_data.dtype, np.uint8):
            if read_data.ndim == 3 and read_data.shape[2] == 3:
                data = _strip_opaque_alpha(cast(npt.NDArray[np.uint8], data))

        return super()._compare_data(data, read_data)


class _ImageIOTgaBackend(_ImageIOFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="TARGA-FI", extension=".tga")
        self.compress = compress

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        flags = 0
        flags |= self.IO_FLAGS.TARGA_SAVE_RLE if self.compress else self.IO_FLAGS.TARGA_DEFAULT
        return flags


class _ImageIOTiffBackend(_ImageIOFreeImageBackend):
    _fp_force_32_bits: ClassVar[
        ImageModifier[ImageDataTypes, ImageDataTypes]
    ] = FPPrecisionModifier(min=32, max=32)

    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(imageio_format="TIFF-FI", extension=".tiff")
        self.compress = compress

    def _get_flags(self, image: Image[ImageDataTypes]) -> int:
        flags = 0
        flags |= self.IO_FLAGS.TIFF_DEFAULT if self.compress else self.IO_FLAGS.TIFF_NONE
        return flags

    def _postprocess(self, image: Image[ImageDataTypes]) -> Image[ImageDataTypes]:
        return self._fp_force_32_bits(image)


def _strip_opaque_alpha(data: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
    if data.ndim == 3 and data.shape[2] == 4 and np.all(data[..., 3] == 255):
        data = data[..., :3]

    return data



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import pathlib
import re

from dataclasses import dataclass
from typing import Generic, Literal, Optional, TypeVar, overload

from no_vtf.image.channel_separator import ChannelSeparator
from no_vtf.image.image import Image, ImageDataTypes
from no_vtf.io.bytes import BytesIO
from no_vtf.io.image import ImageIO
from no_vtf.io.io import IO
from no_vtf.texture.decoder.decoder import TextureDecoder
from no_vtf.texture.extractor.extractor import TextureExtractor
from no_vtf.texture.filter.filter import TextureFilter
from no_vtf.texture.namer.namer import TextureNamer

_A_contra = TypeVar("_A_contra", contravariant=True)
_I = TypeVar("_I", bound=ImageDataTypes)
_T = TypeVar("_T")


@dataclass(frozen=True, kw_only=True)
class Pipeline(Generic[_T, _I]):
    @dataclass(frozen=True, kw_only=True)
    class Receipt:
        io_done: bool

    input_extension_pattern: Optional[re.Pattern[str]] = None

    FORMAT_RAW: Literal["raw"] = "raw"
    FORMAT_SKIP: Literal["skip"] = "skip"
    ldr_format: str
    hdr_format: str

    separate_channels: bool = False

    compress: Optional[bool] = None

    write: Optional[bool] = None
    readback: bool = False

    extractor: TextureExtractor[_T]
    filter: Optional[TextureFilter[_T]]
    decoder: TextureDecoder[_T, _I]
    namer: TextureNamer[_T]

    @classmethod
    def initialize(cls) -> None:
        IO.initialize()

    @overload
    def __call__(self, input_file: pathlib.Path, *, output_file: pathlib.Path) -> Pipeline.Receipt:
        ...

    @overload
    def __call__(
        self, input_file: pathlib.Path, *, output_directory: pathlib.Path
    ) -> Pipeline.Receipt:
        ...

    def __call__(
        self,
        input_file: pathlib.Path,
        *,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
    ) -> Pipeline.Receipt:
        io_done = False

        textures = self.extractor(input_file)

        if self.filter:
            textures = self.filter(textures)

        for texture in textures:
            image = self.decoder(texture)

            image_format = self.hdr_format if image.dynamic_range == "hdr" else self.ldr_format
            if _compare_formats(image_format, self.FORMAT_SKIP):
                continue

            texture_output_file = output_file
            if texture_output_file is None:
                assert output_directory is not None, "output path must be set"

                input_name = input_file.name
                if self.input_extension_pattern:
                    input_name = re.sub(self.input_extension_pattern, "", input_name)

                texture_name = self.namer(input_name, texture)
                texture_output_name = texture_name + "." + image_format
                texture_output_file = output_directory / texture_output_name

            if _compare_formats(image_format, self.FORMAT_RAW):
                assert image.raw, "image must have raw data set"
                io_done = self._do_io(BytesIO(), texture_output_file, image.raw) or io_done
            else:
                image_io = ImageIO(format=image_format, compress=self.compress)
                if not self.separate_channels:
                    io_done = self._do_io(image_io, texture_output_file, image) or io_done
                else:
                    io_done = (
                        self._separate_channels(image_io, texture_output_file, image) or io_done
                    )

        return Pipeline.Receipt(io_done=io_done)

    def _separate_channels(
        self, image_io: ImageIO, path: pathlib.Path, image: Image[ImageDataTypes]
    ) -> bool:
        io_done = False

        channel_separator = ChannelSeparator()
        for image_separated in channel_separator(image):
            new_stem = path.stem + "_" + image_separated.channels
            new_path = path.with_stem(new_stem)
            io_done = self._do_io(image_io, new_path, image_separated) or io_done

        return io_done

    def _do_io(self, io: IO[_A_contra], path: pathlib.Path, data: _A_contra) -> bool:
        io_done = False

        if self.write is not False:
            io_done = self._write(io, path, data) or io_done

        if self.readback:
            io.readback(path, data)
            io_done = True

        return io_done

    def _write(self, io: IO[_A_contra], path: pathlib.Path, data: _A_contra) -> bool:
        assert self.write is not False, "_write() must not be called when writing is disabled"
        skip_existing = self.write is None
        if skip_existing and path.is_file():
            return False

        path.parent.mkdir(parents=True, exist_ok=True)
        io.write(path, data)

        return True


def _compare_formats(a: str, b: str) -> bool:
    return a.lower() == b.lower()



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from abc import abstractmethod
from typing import ClassVar, Generic, TypeVar

import no_vtf.io  # noqa: F401  # define all task runners for IO.initialize()

_A_contra = TypeVar("_A_contra", contravariant=True)


class IO(Generic[_A_contra]):
    _initialized: ClassVar[bool] = False

    @classmethod
    def initialize(cls) -> None:
        for subclass in cls.__subclasses__():
            subclass._initialize()

        cls._initialized = True

    @classmethod
    def _is_initialized(cls) -> bool:
        return cls._initialized

    @classmethod
    def _initialize(cls) -> None:
        pass

    @abstractmethod
    def write(self, path: pathlib.Path, data: _A_contra, /) -> None:
        ...

    @abstractmethod
    def readback(self, path: pathlib.Path, data: _A_contra, /) -> None:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional

from no_vtf.texture.filter.filter import TextureFilter
from no_vtf.texture.vtf import VtfTexture


@dataclass(frozen=True, kw_only=True)
class VtfMipmapFilter(TextureFilter[VtfTexture]):
    mipmap_levels: slice

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if not textures:
            return []

        assert (
            len(set(map(lambda texture: texture.num_mipmaps, textures))) == 1
        ), "num_mipmaps must be the same for all filtered textures"
        indices = range(*self.mipmap_levels.indices(textures[0].num_mipmaps))

        textures = list(filter(lambda texture: texture.mipmap_index in indices, textures))
        return textures


@dataclass(frozen=True, kw_only=True)
class VtfResolutionFilter(TextureFilter[VtfTexture]):
    min: Optional[int] = None
    max: Optional[int] = None

    closest_as_fallback: bool = False

    def __post_init__(self) -> None:
        if self.min is not None and self.max is not None and self.min > self.max:
            raise RuntimeError(f"Unordered resolutions: {self.min = } <= {self.max = }")

    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        if self.min is None and self.max is None:
            return textures

        def resolution_filter(texture: VtfTexture) -> bool:
            if self.min is not None:
                if not all(
                    resolution >= self.min
                    for resolution in (texture.image.logical_width, texture.image.logical_height)
                ):
                    return False

            if self.max is not None:
                if not all(
                    resolution <= self.max
                    for resolution in (texture.image.logical_width, texture.image.logical_height)
                ):
                    return False

            return True

        exact_matches = list(filter(resolution_filter, textures))
        if exact_matches or not self.closest_as_fallback:
            return exact_matches

        assert (
            self.min is not None or self.max is not None
        ), "either min or max resolution must be set"
        num_pixels = (self.min or self.max or 0) * (self.max or self.min or 0)

        close_matches = {
            abs(num_pixels - texture.image.logical_width * texture.image.logical_height): texture
            for texture in textures
        }
        close_matches = dict(sorted(close_matches.items()))

        closest_match = list(close_matches.values())[0:1]
        return closest_match



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import dataclasses

from dataclasses import dataclass
from typing import Optional

from no_vtf.image.image import ImageDynamicRange
from no_vtf.parser.generated.vtf import Vtf as VtfParser


@dataclass(frozen=True, kw_only=True)
class VtfTexture:
    dynamic_range: Optional[ImageDynamicRange]

    is_cubemap: bool

    num_mipmaps: int
    num_frames: int
    num_faces: int
    num_slices: int

    mipmap_index: int
    frame_index: int
    face_index: int
    slice_index: int

    image: VtfParser.Image = dataclasses.field(repr=False, hash=False)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable, Sequence
from typing import ClassVar, Protocol, TypeVar, Union

import no_vtf.task_runner  # noqa: F401  # define all task runners for TaskRunner.initialize()

_A_co = TypeVar("_A_co", covariant=True)


class TaskRunner:
    class Task(Protocol[_A_co]):
        @abstractmethod
        def __call__(self) -> _A_co:
            ...

    _initialized: ClassVar[bool] = False

    @classmethod
    def initialize(cls) -> None:
        for subclass in cls.__subclasses__():
            subclass._initialize()

        cls._initialized = True

    @classmethod
    def _is_initialized(cls) -> bool:
        return cls._initialized

    @classmethod
    def _initialize(cls) -> None:
        pass

    @abstractmethod
    def __call__(
        self, tasks: Sequence[TaskRunner.Task[_A_co]]
    ) -> Iterable[tuple[TaskRunner.Task[_A_co], Union[_A_co, Exception]]]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools
import multiprocessing
import signal

from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from typing import Optional, TypeVar, Union

from no_vtf.task_runner.sequential import SequentialRunner
from no_vtf.task_runner.task_runner import TaskRunner

_A_co = TypeVar("_A_co", covariant=True)


@dataclass(frozen=True, kw_only=True)
class ParallelRunner(TaskRunner):
    max_workers: Optional[int] = None
    initializer: Optional[Callable[[], None]] = None

    @classmethod
    def _initialize(cls) -> None:
        multiprocessing.freeze_support()

    def __post_init__(self) -> None:
        assert self._is_initialized(), "TaskRunner.initialize() must be called early"

    def __call__(
        self, tasks: Sequence[TaskRunner.Task[_A_co]]
    ) -> Iterable[tuple[TaskRunner.Task[_A_co], Union[_A_co, Exception]]]:
        initializer = functools.partial(ParallelRunner._worker_initializer, self.initializer)
        with multiprocessing.Pool(self.max_workers, initializer=initializer) as pool:
            for task, result in pool.imap_unordered(SequentialRunner.process, tasks):
                yield (task, result)

    @staticmethod
    def _worker_initializer(additional_initializer: Optional[Callable[[], None]]) -> None:
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        if additional_initializer:
            additional_initializer()



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Protocol, TypeVar

from no_vtf.image.image import Image, ImageDataTypes

_I_co = TypeVar("_I_co", bound=ImageDataTypes, covariant=True)


class ImageDecoder(Protocol[_I_co]):
    @abstractmethod
    def __call__(
        self, encoded_image: bytes, logical_width: int, logical_height: int, /
    ) -> Image[_I_co]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from collections.abc import Sequence
from typing import Protocol, TypeVar

_T = TypeVar("_T")


class TextureFilter(Protocol[_T]):
    @abstractmethod
    def __call__(self, textures: Sequence[_T]) -> Sequence[_T]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Protocol, TypeVar, Union

from no_vtf.image import Image, ImageDataTypes

_I_co = TypeVar("_I_co", bound=ImageDataTypes, covariant=True)
_I_contra = TypeVar("_I_contra", bound=ImageDataTypes, contravariant=True)


class ImageModifier(Protocol[_I_contra, _I_co]):
    @abstractmethod
    def __call__(self, image: Image[_I_contra], /) -> Image[Union[_I_contra, _I_co]]:
        ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from dataclasses import dataclass
from typing import Literal, Optional, TypeVar, Union

import numpy as np

from no_vtf.image import Image, ImageDataTypes
from no_vtf.image.modifier import ImageModifier

FloatingPointNumBits = Literal[16, 32, 64]

_I_co = TypeVar("_I_co", bound=ImageDataTypes, covariant=True)
_I_contra = TypeVar("_I_contra", bound=ImageDataTypes, contravariant=True)


@dataclass(frozen=True, kw_only=True)
class FPPrecisionModifier(ImageModifier[_I_contra, _I_co]):
    min: Optional[FloatingPointNumBits] = None
    max: Optional[FloatingPointNumBits] = None

    def __post_init__(self) -> None:
        if self.min is not None and self.max is not None and self.min > self.max:
            raise RuntimeError(f"Unordered precisions: {self.min = } <= {self.max = }")

    def __call__(self, image: Image[_I_contra]) -> Image[Union[_I_contra, _I_co]]:
        if not np.issubdtype(image.dtype, np.floating):
            return image

        fp_bits = np.dtype(image.dtype).itemsize * 8

        if self.min is not None and fp_bits < self.min:
            dtype = f"float{self.min}"
            data = image.data.map(lambda data: data.astype(dtype))
            return Image(raw=None, data=data, dtype=dtype, channels=image.channels)

        if self.max is not None and fp_bits > self.max:
            dtype = f"float{self.max}"
            data = image.data.map(lambda data: data.astype(dtype))
            return Image(raw=None, data=data, dtype=dtype, channels=image.channels)

        return image



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from no_vtf.io.io import IO


class BytesIO(IO[bytes]):
    def write(self, path: pathlib.Path, data: bytes) -> None:
        with open(path, "wb") as file:
            file.write(data)

    def readback(self, path: pathlib.Path, data: bytes) -> None:
        with open(path, "rb") as file:
            read_data = file.read()
            if data != read_data:
                raise RuntimeError(f"{path!r}: Data differs from what is in the file")



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from dataclasses import dataclass

from no_vtf.texture.namer.namer import TextureNamer
from no_vtf.texture.vtf import VtfTexture


@dataclass(frozen=True, kw_only=True)
class Vtf2TgaLikeNamer(TextureNamer[VtfTexture]):
    include_mipmap_level: bool

    def __call__(self, input_name: str, texture: VtfTexture) -> str:
        output_name = input_name

        if texture.is_cubemap:
            face_names = ("rt", "lf", "bk", "ft", "up", "dn", "sph")
            output_name += face_names[texture.face_index]

        if texture.num_frames > 1:
            output_name += f"{texture.frame_index:03d}"

        if self.include_mipmap_level and texture.num_mipmaps > 1:
            mipmap_level = texture.num_mipmaps - texture.mipmap_index - 1
            output_name += f"_mip{mipmap_level}"

        if texture.num_slices > 1:
            output_name += f"_z{texture.slice_index:03d}"

        return output_name
